use std::{ops::Deref, sync::Arc};

use anyhow::Context;
use async_trait::async_trait;
use axum::http::{header::SET_COOKIE, HeaderMap};
use oauth2::url::Url;

use crate::{
    introspection::{IdToken, IntrospectionService},
    login::{auth_clap::AuthEngine, config::ConfigClap, AuthClap},
    oauth::{zitadel::ZitadelConfig, OAuth},
    session::{AppSession, SessionService, User},
};

#[async_trait]
pub trait Auth {
    async fn login(&self, return_url: Option<String>) -> anyhow::Result<(HeaderMap, Url)>;
    async fn login_token(&self, user: &str, password: &str) -> anyhow::Result<IdToken>;
    async fn login_authorized(
        &self,
        code: &str,
        state: &str,
        app_session_cookie: Option<String>,
    ) -> anyhow::Result<(HeaderMap, Url)>;
    async fn get_user_from_session(&self, cookie: &str) -> anyhow::Result<User>;
}

#[derive(Clone)]
pub struct AuthService(Arc<dyn Auth + Send + Sync + 'static>);

impl AuthService {
    pub async fn new(config: &AuthClap) -> anyhow::Result<Self> {
        match config.engine {
            AuthEngine::Noop => {
                let session = SessionService::new(config).await?;

                Ok(Self::new_noop(session, &config.config))
            }
            AuthEngine::Zitadel => {
                let session = SessionService::new(config).await?;
                let oauth: OAuth = ZitadelConfig::try_from(config.zitadel.clone())?.into();
                let introspection: IntrospectionService =
                    IntrospectionService::new_zitadel(config).await?;

                Ok(Self::new_zitadel(
                    oauth,
                    introspection,
                    session,
                    &config.config,
                ))
            }
        }
    }

    pub fn new_zitadel(
        oauth: OAuth,
        introspection: IntrospectionService,
        session: SessionService,
        config: &ConfigClap,
    ) -> Self {
        Self(Arc::new(ZitadelAuthService {
            oauth,
            introspection,
            session,
            config: config.clone(),
        }))
    }

    pub fn new_noop(session: SessionService, config: &ConfigClap) -> Self {
        Self(Arc::new(NoopAuthService {
            session,
            config: config.clone(),
        }))
    }
}

impl Deref for AuthService {
    type Target = Arc<dyn Auth + Send + Sync + 'static>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

pub struct ZitadelAuthService {
    oauth: OAuth,
    introspection: IntrospectionService,
    session: SessionService,
    config: ConfigClap,
}
pub static COOKIE_NAME: &str = "SESSION";
pub static COOKIE_APP_SESSION_NAME: &str = "APP_SESSION";

#[async_trait]
impl Auth for ZitadelAuthService {
    async fn login(&self, return_url: Option<String>) -> anyhow::Result<(HeaderMap, Url)> {
        let mut headers = HeaderMap::new();
        if let Some(return_url) = return_url.clone() {
            let cookie_value = self.session.insert(AppSession { return_url }).await?;

            let cookie = format!(
                "{}={}; SameSite=Lax; Path=/",
                COOKIE_APP_SESSION_NAME, cookie_value
            );
            headers.insert(SET_COOKIE, cookie.parse().unwrap());
        }

        let authorize_url = self.oauth.authorize_url().await?;

        Ok((headers, authorize_url))
    }
    async fn login_authorized(
        &self,
        code: &str,
        _state: &str,
        app_session_cookie: Option<String>,
    ) -> anyhow::Result<(HeaderMap, Url)> {
        let token = self.oauth.exchange(code).await?;
        let id_token = self.introspection.get_id_token(token.as_str()).await?;
        let cookie_value = self.session.insert_user("user", id_token).await?;

        let cookie = format!("{}={}; SameSite=Lax; Path=/", COOKIE_NAME, cookie_value);

        let mut headers = HeaderMap::new();
        headers.insert(SET_COOKIE, cookie.parse().unwrap());

        let mut return_url = self.config.return_url.clone();
        if let Some(cookie) = app_session_cookie {
            if let Some(session) = self.session.get(&cookie).await? {
                if session.return_url.starts_with('/') {
                    let mut url = Url::parse(&return_url)?;
                    url.set_path(&session.return_url);
                    return_url = url.to_string();
                } else {
                    return_url = session.return_url;
                }
            }
        }

        Ok((
            headers,
            Url::parse(&return_url)
                .context("failed to parse login_authorized zitadel return url")?,
        ))
    }
    async fn login_token(&self, _user: &str, password: &str) -> anyhow::Result<IdToken> {
        self.introspection.get_id_token(password).await
    }
    async fn get_user_from_session(&self, cookie: &str) -> anyhow::Result<User> {
        match self.session.get_user(cookie).await? {
            Some(u) => Ok(u),
            None => Err(anyhow::anyhow!("failed to find user")),
        }
    }
}

pub struct NoopAuthService {
    session: SessionService,
    config: ConfigClap,
}

#[async_trait]
impl Auth for NoopAuthService {
    async fn login(&self, return_url: Option<String>) -> anyhow::Result<(HeaderMap, Url)> {
        let url = Url::parse(&format!(
            "{}/auth/authorized?code=noop&state=noop",
            self.config
                .return_url
                .rsplit_once('/')
                .map(|(a, _)| a)
                .unwrap()
        ))
        .unwrap();
        Ok((HeaderMap::new(), url))
    }
    async fn login_authorized(
        &self,
        _code: &str,
        _state: &str,
        _app_session_cookie: Option<String>,
    ) -> anyhow::Result<(HeaderMap, Url)> {
        let cookie_value = self
            .session
            .insert_user(
                "user",
                IdToken {
                    sub: uuid::Uuid::new_v4().to_string(),
                    email: format!("{}@kjuulh.io", uuid::Uuid::new_v4()),
                    name: uuid::Uuid::new_v4().to_string(),
                },
            )
            .await?;

        let cookie = format!("{}={}; SameSite=Lax; Path=/", COOKIE_NAME, cookie_value);

        let mut headers = HeaderMap::new();
        headers.insert(SET_COOKIE, cookie.parse().unwrap());

        Ok((
            headers,
            Url::parse(&self.config.return_url)
                .context("failed to parse login_authorized zitadel return url")?,
        ))
    }

    async fn login_token(&self, _user: &str, _password: &str) -> anyhow::Result<IdToken> {
        todo!()
    }

    async fn get_user_from_session(&self, cookie: &str) -> anyhow::Result<User> {
        match self.session.get_user(cookie).await? {
            Some(u) => Ok(u),
            None => Err(anyhow::anyhow!("failed to find user")),
        }
    }
}
